unit FMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, IBDatabase, DB, IBCustomDataSet, IBQuery, IBSQL;

type
  TfmMain = class(TForm)
    lbSourceOwner: TLabel;
    edSourceOwner: TEdit;
    lbSourcePassword: TLabel;
    edSourcePassword: TEdit;
    lbDestinationOwner: TLabel;
    edDestinationOwner: TEdit;
    lbDestinationPassword: TLabel;
    edDestinationPassword: TEdit;
    ibdMain: TIBDatabase;
    ibtMain: TIBTransaction;
    lbDatabase: TLabel;
    edDatabase: TEdit;
    bbMigratePrivileges: TButton;
    ibqMain: TIBQuery;
    ibsMain: TIBSQL;
    procedure bbMigratePrivilegesClick(Sender: TObject);
  private
    { Private-Deklarationen }
    function ConnectToDatabase(ADatabaseName, AUser, APassword: String): Boolean;
    procedure GrantPrivileges;
    procedure MigrateExistingPrivileges;
  public
    { Public-Deklarationen }
  end;

var
  fmMain: TfmMain;

implementation

{$R *.dfm}

{ TForm1 }

function TfmMain.ConnectToDatabase(ADatabaseName, AUser, APassword: String): Boolean;
begin
  with ibdMain do
  begin
    Connected := False;
    LoginPrompt := False;
    DatabaseName := ADatabaseName;
    Params.Clear;
    Params.Add('user_name=' + AUser);
    Params.Add('password=' + APassword);
    Connected := True;
    Result := Connected;
  end;
end;

procedure TfmMain.GrantPrivileges;
var
  relationName, procedureName: String;
begin
  // Connect with source owner
  if ConnectToDatabase(edDatabase.Text, edSourceOwner.Text, edSourcePassword.Text) then
  begin
    ibtMain.StartTransaction;
    try
      // Grant permissions for tables and views
      with ibqMain do
      begin
        Close;
        SQL.Text := 'select rdb$relation_name from rdb$relations where (rdb$system_flag is null or rdb$system_flag = 0)';
        Open;
        while not EOF do
        begin
          relationName := FieldByName('rdb$relation_name').AsString;
          ibsMain.SQL.Text := Format('REVOKE ALL ON %s FROM %s', [relationName, edDestinationOwner.Text]);
          ibsMain.ExecQuery;
          ibsMain.SQL.Text := Format('GRANT ALL ON %s TO %s WITH GRANT OPTION', [relationName, edDestinationOwner.Text]);
          ibsMain.ExecQuery;
          Next;
        end;
      end;
      // Grant permissions for stored procedures
      with ibqMain do
      begin
        Close;
        SQL.Text := 'select rdb$procedure_name from rdb$procedures where (rdb$system_flag is null or rdb$system_flag = 0)';
        Open;
        while not EOF do
        begin
          procedureName := FieldByName('rdb$procedure_name').AsString;
          ibsMain.SQL.Text := Format('REVOKE EXECUTE ON PROCEDURE %s FROM %s', [procedureName, edDestinationOwner.Text]);
          ibsMain.ExecQuery;
          ibsMain.SQL.Text := Format('GRANT EXECUTE ON PROCEDURE %s TO %s WITH GRANT OPTION', [procedureName, edDestinationOwner.Text]);
          ibsMain.ExecQuery;
          Next;
        end;
      end;
      ibtMain.Commit
    except
      ibtMain.Rollback;
      raise;
    end;
  end;
end;

procedure TfmMain.MigrateExistingPrivileges;
var
  userType, grantOption: Smallint;
  privilege: Char;
  grantStmt, relationName, userName, fieldName, userTypeName: String;
begin
  // Connect with destination owner
  if ConnectToDatabase(edDatabase.Text, edDestinationOwner.Text, edDestinationPassword.Text) then
  begin
    ibtMain.StartTransaction;
    try
      with ibqMain do
      begin
        Close;
        SQL.Text := 'SELECT * FROM RDB$USER_PRIVILEGES WHERE RDB$GRANTOR=:grantor ORDER BY RDB$RELATION_NAME';
        if not Prepared then Prepare;
        ParamByName('grantor').AsString := edSourceOwner.Text;
        Open;
        while not EOF do
        begin
          grantStmt := '';
          privilege := FieldByName('RDB$PRIVILEGE').AsString[1];
          relationName := FieldByName('RDB$RELATION_NAME').AsString;
          userName := FieldByName('RDB$USER').AsString;
          grantOption := FieldByName('RDB$GRANT_OPTION').AsInteger;
          fieldName := FieldByName('RDB$FIELD_NAME').AsString;
          userType := FieldByName('RDB$USER_TYPE').AsInteger;
          case privilege of
            'A': begin // ALL
              grantStmt := Format('GRANT ALL ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'S': begin // SELECT
              grantStmt := Format('GRANT SELECT ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'D': begin // DELETE
              grantStmt := Format('GRANT DELETE ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'I': begin // INSERT
              grantStmt := Format('GRANT INSERT ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'U': begin // UPDATE
              grantStmt := 'GRANT UPDATE ';
              if fieldName <> '' then grantStmt := grantStmt + '(' + fieldName + ')';
              grantStmt := grantStmt + Format(' ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'R': begin // REFERENCE
              grantStmt := Format('GRANT REFERENCES ON %s TO %s', [relationName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
            'M': begin // MEMBER OF
              grantStmt := Format('GRANT %s TO %s', [relationName, userName]);
              if grantOption = 2 then grantStmt := grantStmt + ' WITH ADMIN OPTION';
            end;
            'X': begin // EXECUTE
              userTypeName := '';
              case userType of
                1: userTypeName := 'VIEW';
                2: userTypeName := 'TRIGGER';
                5: userTypeName := 'PROCEDURE';
                8: userTypeName := 'USER';
              end;
              grantStmt := Format('GRANT EXECUTE ON PROCEDURE %s TO %s %s', [relationName, userTypeName, userName]);
              if grantOption = 1 then grantStmt := grantStmt + ' WITH GRANT OPTION';
            end;
          end;
          ibsMain.SQL.Text := grantStmt;
          ibsMain.ExecQuery;
          Next;
        end;
      end;
      ibtMain.Commit;
    except
      ibtMain.Rollback;
      raise;
    end;
  end;
end;

procedure TfmMain.bbMigratePrivilegesClick(Sender: TObject);
begin
  if MessageDlg('Migrate SQL privileges', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
  begin
    GrantPrivileges;
    MigrateExistingPrivileges;
    ShowMessage('Migrate SQL privileges succeeded!');
  end;
end;

end.
